#include <iostream>
#include <seplib.h>
#include <stdio.h>
#include <cmath>

int main(int argc, char* argv[])
{
	/*initialize IO*/
	initpar(argc,argv);
	
	/* declare variables*/
	int 	n1, n2, m1, m2; 
	float	d1,d2;
	float 	*d, *f, *b;
	float 	beta, mu;
	char	back_out[128];

	int 	tr, k, off;
	float 	ftemp;
	float	*K, *C, *D, *ff, *bb, *btemp;

	/*retrieve header info and read in data*/	
	//Get values from header file
	if(!hetch("n1","d",&n1)) seperr("could not get n1 from file\n");
	if(!hetch("n2","d",&n2)) seperr("could not get n2 from file\n");
	if(!hetch("d1","f",&d1)) seperr("could not get d1 from file\n");
	if(!hetch("d2","f",&d2)) seperr("could not get d2 from file\n");
	if(!getch("m1","d",&m1)) seperr("No lattice stage number 1! \n");
	if(!getch("m2","d",&m2)) seperr("No lattice stage number 2! \n");
	if(!getch("beta","f",&beta)) beta = 1;
	if(!getch("mu","f",&mu)) mu = 1-beta;
	if(!getch("back","s",&back_out)) seperr("Backward error file not defined\n");

	// INPUT
	d = new float[n1*n2];
	if(!sreed("in",d,sizeof(*d)*n1*n2)) seperr("could not read data from std in file\n");
	// OUTPUT
	f = new float[n1*n2];
	b = new float[n1*n2];
	// AUXILIARY
	//int m = m1*m2-1;
	int m = m1*(2*m2-1)-m2;
	int nK = m*(m+1)/2;
	C = new float[n1*nK];
	D = new float[n1*nK];
	K = new float[n1*nK];

	// FOR 2d lattice
	ff = new float[m];
	bb = new float[m];
	btemp = new float[m];
		
	for (int im = 0; im < n1*nK; im++){
		C[im] = 0; 
		D[im] = 1e-3;
		K[im] = 0;
	}	

	// loop over traces 
	int count1 = 0;
	int count2 = 0;
	int t;
	for (int i2 = m2-1; i2 < n2; i2++){
		tr = i2*n1;
		// loop over time
		for (int i1 = m1-1; i1 < n1-m1+1; i1++) {

			count1=0;
			count2=0;

			k = 1;
			// ordering to 1D array (asymmetric)
			int jj=1;
			for (int j2 = 0; j2 < m2; j2++) {

				//int jj = (j2 > 0 ? 0 : 1);
				int kk = 1-jj;

				off = (i2-j2)*n1 + i1+(m1-1)*kk-2*(m1-1)*(j2%2);				

				for (int j1 = 0; j1 < m1-(j2+1)/m2+(m1-1)*kk; j1++) {
					ff[count1] = d[off- k*j1];
					count1+=1;
				}				
				
				for (int j1 = 0; j1 < m1-jj+(m1-1)*kk; j1++) {
					//bb[j2*m1+j1-(1-jj)] = d[off - k*j1 - k*jj];
					//btemp[j2*m1+j1-(1-jj)] = d[off - k*j1 - k*jj];
					btemp[count2] = d[off - k*j1 - jj];
					count2+=1;
				}
				// switch direction of ordering
				k *= -1; 
				jj=0;
			}

			// loop over lattice stages
			k = 0;

			for (int it = 0; it < m; it++) {
				for (int im = 0; im < m-it; im++) {
					// Burg's estimate
					//C[k+i1*nK] = beta1*C[k+i1*nK] + 2*ff[im]*btemp[im+it];			
					D[k+i1*nK] = beta*D[k+(i1-1)*nK] + (1-beta)*(ff[im]*ff[im] + btemp[im+it]*btemp[im+it]);
					//D[k+i1*nK] = 1/beta1*D[k+i1*nK]*(1-(1-beta1)/beta1*D[k+i1*nK]*(ff[im]*ff[im] + btemp[im+it]*btemp[im+it]));
					//if (D[k] < 1e-10) D[k] += 1e-10;

					// forward and backward errors
					//K = -C[k+i1*nK]/D[k+i1*nK];
					ftemp = ff[im];
					ff[im] = ftemp + K[k+i1*nK]*btemp[im+it];
					bb[im+it] = btemp[im+it] + K[k+i1*nK]*ftemp;
					
					K[k+(i1+1)*nK] = K[k+i1*nK] - mu*(ftemp*bb[im+it] + btemp[im+it]*ff[im])/D[k+i1*nK];

					btemp[im+it] = bb[im+it];
					k += 1;
				}

			}
			
			
			// update fwd and backward error
			f[i1+tr] = ff[0];
			b[i1+tr] = bb[m-1];
		}
		
		// update going to the next trace
		// for (int ik=0; ik < nK*n1; ik++) {
		// 		D[ik] *= beta2;
		// 		//K[ik] *= beta2;
		// 		//C[ik] *= beta2;
		// 	}
			
	}

	// OUTPUT
	if(!srite("out",f,sizeof*f*n1*n2)) seperr("Could not write Std_out data");		
	if(putch("n1","d",&n1)) seperr("Could not write n1 to file");
	if(putch("n2","d",&n2)) seperr("Could not write n2 to file");
	if(putch("d1","f",&d1)) seperr("Could not write d1 to file");
	if(putch("d2","f",&d2)) seperr("Could not write d1 to file");

	if(!srite(back_out,b,sizeof(*b)*n1*n2)) seperr("Could not write Backward error");	
	if(auxputch("n1","d",&n1, back_out)) seperr("Could not write n1 to back_out");
	if(auxputch("n2","d",&n2, back_out)) seperr("Could not write n2 to back_out");
	if(auxputch("d1","f",&d1, back_out)) seperr("Could not write d1 to back_out");
	if(auxputch("d2","f",&d2, back_out)) seperr("Could not write d2 to back_out");

	//CLEANUP
	delete[] d;
	delete[] f;
	delete[] b;
	delete[] C;
	delete[] D;
	delete[] K;
	delete[] ff;
	delete[] bb;
	delete[] btemp;

	return 0;
}

