#include <iostream>
#include <seplib.h>
#include <stdio.h>
#include <cmath>

int main(int argc, char* argv[])
{
	/*initialize IO*/
	initpar(argc,argv);
	
	/* declare variables*/
	int 	n1, n2, m1, m2, gap; 
	float	d1,d2;
	float	delta;
	float	lambda;
	float	*d,*e;
	
	int 	off, count1;
	float	uPu, err;
	float	*w,*P,*k;
	float	*dd,*Pu,*kuP;

//////////////////////////////////////////////////////////
	/*retrieve header info and read in data*/	
	//Get values from header file
	if(!hetch("n1","d",&n1)) seperr("could not get n1 from file\n");
	if(!hetch("n2","d",&n2)) seperr("could not get n2 from file\n");
	if(!hetch("d1","f",&d1)) seperr("could not get d1 from file\n");
	if(!hetch("d2","f",&d2)) seperr("could not get d2 from file\n");
	if(!getch("m1","d",&m1)) seperr("No m1! \n");
	if(!getch("m2","d",&m2)) seperr("No m2! \n");
	if(!getch("delta","f",&delta)) delta=0.001;
	if(!getch("lambda","f",&lambda)) lambda=1;
	if(!getch("gap","d",&gap)) gap=1;

	fprintf(stderr, "RLS filter with m1=%d m2=%d delta=%f lambda=%f gap=%d\n", m1,m2,delta,lambda,gap);

	// INPUT
	d = new float[n1*n2];
	if(!sreed("in",d,sizeof(*d)*n1*n2)) seperr("could not read data from std in file\n");
	// OUTPUT
	e = new float[n1*n2];

//////////////////////////////////////////////////////////
	// AUXILIARY

	// filter
	int m = m1*m2+(m1-1)*(m2-1);
	w = new float[m];
	//gain factor
	k = new float[m];
	//inverse of aurocorr matrix
	P = new float[m*m];
	//local 1d array
	dd = new float[m];
	// P(n-1)u(n)
	Pu = new float[m];
	// aux
	kuP = new float[m*m];
	
	
//////////////////////////////////////////////////////////
	
	for (int i2 = m2-1; i2 < n2; i2++){
		
		//initialize
		for (int i=0; i<m; i++) {
			w[i] = 0;
		}
		for (int i=0; i<m*m; i++) {
			P[i] = 0;
		}
		for (int i=0; i<m; i++) {
			P[i+i*m] = 1/delta;
		}

		
		// loop over time
		for (int i1 = m1-1; i1 < n1-m1+1-gap; i1++) {
			
			err = 0;
			count1=0;
			uPu = 0;
			for (int i=0; i<m; i++) {
				Pu[i] = 0;
			}
			for (int i=0; i<m*m; i++) {
				kuP[i] = 0;
			}

			int _k = 1;
			// ordering to 1D array (asymmetric)
			int kk=0;
			for (int j2 = 0; j2 < m2; j2++) {

				off = (i2-j2)*n1 + i1+(m1-1)*kk-2*(m1-1)*(j2%2);				

				for (int j1 = 0; j1 < m1+(m1-1)*kk; j1++) {
					dd[count1] = d[off- _k*j1];
					count1+=1;
				}								
				// switch direction of ordering
				_k *= -1; 
				kk=1;
			}


		for (int im=0; im<m; im++) {
			for (int ik=0; ik<m; ik++) {
				Pu[im] += P[ik+im*m]*dd[ik];
			}
			uPu += dd[im]*Pu[im];
		}

		for (int im=0; im<m; im++) {
			k[im] = (1/lambda) * Pu[im]/(1+(1/lambda) * uPu);
			err -= w[im]*dd[im];
		}
		// error of future sample
		err += d[i1+i2*n1+gap];
		// update filter
		for (int im=0; im<m; im++) {
			w[im] += err*k[im];
		}
		// update error
		err=0;
		for (int im=0; im<m; im++) {
			err -= w[im]*dd[im];
		}
		err += d[i1+i2*n1+gap];
		// update P
		for (int i=0; i< m; i++) {
			for (int im=0; im<m; im++) {
				for (int ik=0; ik<m; ik++) {
					kuP[i+im*m] -= 1/lambda*k[im]*dd[ik]*P[i+ik*m];		
				}
			}
		}
		for (int im=0; im<m; im++) {
			for (int ik=0; ik<m; ik++) {
				P[ik+im*m] = 1/lambda*P[ik+im*m] + kuP[ik+im*m];
			}
		}
		e[i1+i2*n1] = err;
		
		}	
	}

	// OUTPUT
	if(!srite("out",e,sizeof*e*n1*n2)) seperr("Could not write Std_out data");		
	if(putch("n1","d",&n1)) seperr("Could not write n1 to file");
	if(putch("n2","d",&n2)) seperr("Could not write n2 to file");
	if(putch("d1","f",&d1)) seperr("Could not write d1 to file");
	if(putch("d2","f",&d2)) seperr("Could not write d1 to file");

	//CLEANUP
	delete[] d;
	delete[] dd;
	delete[] Pu;
	delete[] kuP;
	delete[] e;
	delete[] w;
	delete[] k;
	delete[] P;

	return 0;
}

